# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the regressors for scikit-learn models.
"""
import logging
import os
import pickle
from copy import deepcopy
from typing import List, Optional, Tuple, Union, TYPE_CHECKING

import numpy as np

from art.estimators.scikitlearn import ScikitlearnEstimator
from art.estimators.regression.regressor import RegressorMixin
from art import config

if TYPE_CHECKING:
    # pylint: disable=C0412
    import sklearn

    from art.utils import CLIP_VALUES_TYPE, PREPROCESSING_TYPE
    from art.defences.preprocessor import Preprocessor
    from art.defences.postprocessor import Postprocessor
    from art.metrics.verification_decisions_trees import LeafNode

logger = logging.getLogger(__name__)


class ScikitlearnRegressor(RegressorMixin, ScikitlearnEstimator):
    """
    Wrapper class for scikit-learn regression models.
    """

    estimator_params = ScikitlearnEstimator.estimator_params

    def __init__(
        self,
        model: "sklearn.base.BaseEstimator",
        clip_values: Optional["CLIP_VALUES_TYPE"] = None,
        preprocessing_defences: Union["Preprocessor", List["Preprocessor"], None] = None,
        postprocessing_defences: Union["Postprocessor", List["Postprocessor"], None] = None,
        preprocessing: "PREPROCESSING_TYPE" = (0.0, 1.0),
    ) -> None:
        """
        Create a `Regressor` instance from a scikit-learn regressor model.

        :param model: scikit-learn regressor model.
        :param clip_values: Tuple of the form `(min, max)` representing the minimum and maximum values allowed
               for features.
        :param preprocessing_defences: Preprocessing defence(s) to be applied by the classifier.
        :param postprocessing_defences: Postprocessing defence(s) to be applied by the classifier.
        :param preprocessing: Tuple of the form `(subtrahend, divisor)` of floats or `np.ndarray` of values to be
               used for data preprocessing. The first value will be subtracted from the input. The input will then
               be divided by the second one.
        """
        super().__init__(
            model=model,
            clip_values=clip_values,
            preprocessing_defences=preprocessing_defences,
            postprocessing_defences=postprocessing_defences,
            preprocessing=preprocessing,
        )
        self._input_shape = self._get_input_shape(model)

    @property
    def input_shape(self) -> Tuple[int, ...]:
        """
        Return the shape of one input sample.

        :return: Shape of one input sample.
        """
        return self._input_shape  # type: ignore

    def fit(self, x: np.ndarray, y: np.ndarray, **kwargs) -> None:
        """
        Fit the regressor on the training set `(x, y)`.

        :param x: Training data.
        :param y: Target values.
        :param kwargs: Dictionary of framework-specific arguments. These should be parameters supported by the
               `fit` function in `sklearn` regressor and will be passed to this function as such.
        """
        # Apply preprocessing
        x_preprocessed, y_preprocessed = self._apply_preprocessing(x, y, fit=True)

        self.model.fit(x_preprocessed, y_preprocessed, **kwargs)
        self._input_shape = self._get_input_shape(self.model)

    def predict(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """
        Perform prediction for a batch of inputs.

        :param x: Input samples.
        :return: Array of predictions.
        :raises `ValueError`: If the regressor does not have the method `predict`
        """
        # Apply defences
        x_preprocessed, _ = self._apply_preprocessing(x, y=None, fit=False)

        if callable(getattr(self.model, "predict", None)):
            y_pred = self.model.predict(x_preprocessed)
        else:
            raise ValueError("The provided model does not have the method `predict`.")

        # Apply postprocessing
        predictions = self._apply_postprocessing(preds=y_pred, fit=False)

        return predictions

    def save(self, filename: str, path: Optional[str] = None) -> None:
        """
        Save a model to file in the format specific to the backend framework.

        :param filename: Name of the file where to store the model.
        :param path: Path of the folder where to store the model. If no path is specified, the model will be stored in
                     the default data location of the library `ART_DATA_PATH`.
        """
        if path is None:
            full_path = os.path.join(config.ART_DATA_PATH, filename)
        else:
            full_path = os.path.join(path, filename)
        folder = os.path.split(full_path)[0]
        if not os.path.exists(folder):
            os.makedirs(folder)

        with open(full_path + ".pickle", "wb") as file_pickle:
            pickle.dump(self.model, file=file_pickle)

    def clone_for_refitting(self) -> "ScikitlearnRegressor":
        """
        Create a copy of the classifier that can be refit from scratch.

        :return: new estimator
        """
        import sklearn

        clone = type(self)(sklearn.base.clone(self.model))
        params = self.get_params()
        del params["model"]
        clone.set_params(**params)
        return clone

    def reset(self) -> None:
        """
        Resets the weights of the classifier so that it can be refit from scratch.

        """
        # No need to do anything since scikitlearn models start from scratch each time fit() is called
        pass

    def compute_loss(self, x: np.ndarray, y: np.ndarray, **kwargs) -> np.ndarray:
        """
        Compute the MSE loss of the regressor for samples `x`.

        :param x: Input samples.
        :param y: Target values.
        :return: Loss values.
        """

        return (y - self.predict(x)) ** 2

    def compute_loss_from_predictions(self, pred: np.ndarray, y: np.ndarray, **kwargs) -> np.ndarray:
        """
        Compute the MSE loss of the regressor for predictions `pred`.

        :param pred: Model predictions.
        :param y: Target values.
        :return: Loss values.
        """

        return (y - pred) ** 2


class ScikitlearnDecisionTreeRegressor(ScikitlearnRegressor):
    """
    Wrapper class for scikit-learn Decision Tree Regressor models.
    """

    def __init__(
        self,
        model: "sklearn.tree.DecisionTreeRegressor",
        clip_values: Optional["CLIP_VALUES_TYPE"] = None,
        preprocessing_defences: Union["Preprocessor", List["Preprocessor"], None] = None,
        postprocessing_defences: Union["Postprocessor", List["Postprocessor"], None] = None,
        preprocessing: "PREPROCESSING_TYPE" = (0.0, 1.0),
    ) -> None:
        """
        Create a `Regressor` instance from a scikit-learn Decision Tree Regressor model.

        :param model: scikit-learn Decision Tree Regressor model.
        :param clip_values: Tuple of the form `(min, max)` representing the minimum and maximum values allowed
               for features.
        :param preprocessing_defences: Preprocessing defence(s) to be applied by the classifier.
        :param postprocessing_defences: Postprocessing defence(s) to be applied by the classifier.
        :param preprocessing: Tuple of the form `(subtrahend, divisor)` of floats or `np.ndarray` of values to be
               used for data preprocessing. The first value will be subtracted from the input. The input will then
               be divided by the second one.
        """
        # pylint: disable=E0001
        import sklearn

        if not isinstance(model, sklearn.tree.DecisionTreeRegressor):
            raise TypeError("Model must be of type sklearn.tree.DecisionTreeRegressor.")

        super().__init__(
            model=model,
            clip_values=clip_values,
            preprocessing_defences=preprocessing_defences,
            postprocessing_defences=postprocessing_defences,
            preprocessing=preprocessing,
        )
        self._model = model

    def get_values_at_node(self, node_id: int) -> np.ndarray:
        """
        Returns the feature of given id for a node.

        :return: Normalized values at node node_id.
        """
        return self.model.tree_.value[node_id]

    def get_left_child(self, node_id: int) -> int:
        """
        Returns the id of the left child node of node_id.

        :return: The indices of the left child in the tree.
        """
        return self.model.tree_.children_left[node_id]

    def get_right_child(self, node_id: int) -> int:
        """
        Returns the id of the right child node of node_id.

        :return: The indices of the right child in the tree.
        """
        return self.model.tree_.children_right[node_id]

    def get_decision_path(self, x: np.ndarray) -> np.ndarray:
        """
        Returns the path through nodes in the tree when classifying x. Last one is leaf, first one root node.

        :return: The indices of the nodes in the array structure of the tree.
        """
        if len(np.shape(x)) == 1:
            return self.model.decision_path(x.reshape(1, -1)).indices

        return self.model.decision_path(x).indices

    def get_threshold_at_node(self, node_id: int) -> float:
        """
        Returns the threshold of given id for a node.

        :return: Threshold value of feature split in this node.
        """
        return self.model.tree_.threshold[node_id]

    def get_feature_at_node(self, node_id: int) -> int:
        """
        Returns the feature of given id for a node.

        :return: Feature index of feature split in this node.
        """
        return self.model.tree_.feature[node_id]

    def get_samples_at_node(self, node_id: int) -> int:
        """
        Returns the number of training samples mapped to a node.

        :return: Number of samples mapped this node.
        """
        return self.model.tree_.n_node_samples[node_id]

    def _get_leaf_nodes(self, node_id, i_tree, class_label, box) -> List["LeafNode"]:
        from art.metrics.verification_decisions_trees import LeafNode, Box, Interval

        leaf_nodes: List[LeafNode] = []

        if self.get_left_child(node_id) != self.get_right_child(node_id):

            node_left = self.get_left_child(node_id)
            node_right = self.get_right_child(node_id)

            box_left = deepcopy(box)
            box_right = deepcopy(box)

            feature = self.get_feature_at_node(node_id)
            box_split_left = Box(intervals={feature: Interval(-np.inf, self.get_threshold_at_node(node_id))})
            box_split_right = Box(intervals={feature: Interval(self.get_threshold_at_node(node_id), np.inf)})

            if box.intervals:
                box_left.intersect_with_box(box_split_left)
                box_right.intersect_with_box(box_split_right)
            else:
                box_left = box_split_left
                box_right = box_split_right

            leaf_nodes += self._get_leaf_nodes(node_left, i_tree, class_label, box_left)
            leaf_nodes += self._get_leaf_nodes(node_right, i_tree, class_label, box_right)

        else:
            leaf_nodes.append(
                LeafNode(
                    tree_id=i_tree,
                    class_label=class_label,
                    node_id=node_id,
                    box=box,
                    value=self.get_values_at_node(node_id)[0, 0],
                )
            )

        return leaf_nodes



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging

import numpy as np
import cv2
import pytest

from art.config import ART_NUMPY_DTYPE
from art.preprocessing.image import ImageResize, ImageResizePyTorch, ImageResizeTensorFlowV2
from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.fixture()
def image_batch(channels_first, label_type):
    channels = 3
    height = 20
    width = 16

    if channels_first:
        data_shape = (2, channels, height, width)
    else:
        data_shape = (2, height, width, channels)
    x = (0.5 * np.ones(data_shape)).astype(ART_NUMPY_DTYPE)

    if label_type == "classification":
        y = np.arange(len(x))
    elif label_type == "object_detection":
        y = []
        for _ in range(len(x)):
            y_1, x_1 = np.random.uniform(0, (height / 2, width / 2))
            y_2, x_2 = np.random.uniform((y_1 + 3, x_1 + 3), (height, width))
            target_dict = {
                "boxes": np.array([[x_1, y_1, x_2, y_2]]),
                "labels": np.array([0]),
            }
            y.append(target_dict)
    else:
        y = None

    return x, y


@pytest.mark.framework_agnostic
@pytest.mark.parametrize("height", [16, 20, 24])
@pytest.mark.parametrize("width", [16, 20, 24])
@pytest.mark.parametrize("channels_first", [True, False])
@pytest.mark.parametrize("label_type", [None, "classification", "object_detection"])
@pytest.mark.parametrize("interpolation", [cv2.INTER_LINEAR, cv2.INTER_AREA])
def test_resize_numpy(height, width, channels_first, label_type, interpolation, image_batch, art_warning):
    x, y = image_batch

    try:
        resize = ImageResize(
            height=height,
            width=width,
            channels_first=channels_first,
            label_type=label_type,
            interpolation=interpolation,
        )

        x_preprocess, y_preprocess = resize(x, y)

        if channels_first:
            assert x_preprocess.shape == (x.shape[0], x.shape[1], height, width)
        else:
            assert x_preprocess.shape == (x.shape[0], height, width, x.shape[3])

        if label_type == "classification":
            np.testing.assert_array_equal(y, y_preprocess)
        elif label_type == "object_detection":
            assert y[0]["boxes"].shape == y_preprocess[0]["boxes"].shape
            np.testing.assert_array_equal(y[0]["labels"], y_preprocess[0]["labels"])
        else:
            assert y is None

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
@pytest.mark.parametrize("height", [16, 20, 24])
@pytest.mark.parametrize("width", [16, 20, 24])
@pytest.mark.parametrize("channels_first", [True, False])
@pytest.mark.parametrize("label_type", [None, "classification", "object_detection"])
@pytest.mark.parametrize("interpolation", ["bilinear", "area"])
def test_resize_pytorch(height, width, channels_first, label_type, interpolation, image_batch, art_warning):
    import torch

    x, y = image_batch

    x = torch.from_numpy(x)
    if label_type == "classification":
        y = torch.from_numpy(y)
    elif label_type == "object_detection":
        y = [{k: torch.from_numpy(v) for k, v in y_i.items()} for y_i in y]

    try:
        resize = ImageResizePyTorch(
            height=height,
            width=width,
            channels_first=channels_first,
            label_type=label_type,
            interpolation=interpolation,
        )

        x_preprocess, y_preprocess = resize.forward(x, y)

        if channels_first:
            assert x_preprocess.shape == (x.shape[0], x.shape[1], height, width)
        else:
            assert x_preprocess.shape == (x.shape[0], height, width, x.shape[3])

        if label_type == "classification":
            np.testing.assert_array_equal(y, y_preprocess)
        elif label_type == "object_detection":
            assert y[0]["boxes"].shape == y_preprocess[0]["boxes"].shape
            np.testing.assert_array_equal(y[0]["labels"], y_preprocess[0]["labels"])
        else:
            assert y is None

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
@pytest.mark.parametrize("height", [16, 20, 24])
@pytest.mark.parametrize("width", [16, 20, 24])
@pytest.mark.parametrize("channels_first", [True, False])
@pytest.mark.parametrize("label_type", [None, "classification", "object_detection"])
@pytest.mark.parametrize("interpolation", ["bilinear", "area"])
def test_resize_tensorflow(height, width, channels_first, label_type, interpolation, image_batch, art_warning):
    import tensorflow as tf

    x, y = image_batch

    x = tf.convert_to_tensor(x)
    if label_type == "classification":
        y = tf.convert_to_tensor(y)
    elif label_type == "object_detection":
        y = [{k: tf.convert_to_tensor(v) for k, v in y_i.items()} for y_i in y]

    try:
        resize = ImageResizeTensorFlowV2(
            height=height,
            width=width,
            channels_first=channels_first,
            label_type=label_type,
            interpolation=interpolation,
        )

        x_preprocess, y_preprocess = resize.forward(x, y)

        if channels_first:
            assert x_preprocess.shape == (x.shape[0], x.shape[1], height, width)
        else:
            assert x_preprocess.shape == (x.shape[0], height, width, x.shape[3])

        if label_type == "classification":
            np.testing.assert_array_equal(y, y_preprocess)
        elif label_type == "object_detection":
            assert y[0]["boxes"].shape == y_preprocess[0]["boxes"].shape
            np.testing.assert_array_equal(y[0]["labels"], y_preprocess[0]["labels"])
        else:
            assert y is None

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.framework_agnostic
def test_check_params_numpy(art_warning):
    try:
        with pytest.raises(ValueError):
            _ = ImageResize(height=0, width=32)

        with pytest.raises(ValueError):
            _ = ImageResize(height=32, width=0)

        with pytest.raises(ValueError):
            _ = ImageResize(height=0, width=0)

        with pytest.raises(ValueError):
            _ = ImageResize(height=32, width=32, clip_values=(0,))

        with pytest.raises(ValueError):
            _ = ImageResize(height=32, width=32, clip_values=(1, 0))

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("pytorch")
def test_check_params_pytorch(art_warning):
    try:
        with pytest.raises(ValueError):
            _ = ImageResizePyTorch(height=0, width=32)

        with pytest.raises(ValueError):
            _ = ImageResizePyTorch(height=32, width=0)

        with pytest.raises(ValueError):
            _ = ImageResizePyTorch(height=0, width=0)

        with pytest.raises(ValueError):
            _ = ImageResizePyTorch(height=32, width=32, clip_values=(0,))

        with pytest.raises(ValueError):
            _ = ImageResizePyTorch(height=32, width=32, clip_values=(1, 0))

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.only_with_platform("tensorflow2")
def test_check_params_tensorflow(art_warning):
    try:
        with pytest.raises(ValueError):
            _ = ImageResizeTensorFlowV2(height=0, width=32)

        with pytest.raises(ValueError):
            _ = ImageResizeTensorFlowV2(height=32, width=0)

        with pytest.raises(ValueError):
            _ = ImageResizeTensorFlowV2(height=0, width=0)

        with pytest.raises(ValueError):
            _ = ImageResizeTensorFlowV2(height=32, width=32, clip_values=(0,))

        with pytest.raises(ValueError):
            _ = ImageResizeTensorFlowV2(height=32, width=32, clip_values=(1, 0))

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Creates a Deep Partition Aggregation ensemble classifier.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import warnings
from typing import List, Optional, Union, Callable, Dict, TYPE_CHECKING

import copy
import numpy as np

from art.estimators.classification.ensemble import EnsembleClassifier

if TYPE_CHECKING:
    from art.utils import CLIP_VALUES_TYPE, PREPROCESSING_TYPE, CLASSIFIER_NEURALNETWORK_TYPE
    from art.defences.preprocessor import Preprocessor
    from art.defences.postprocessor import Postprocessor

logger = logging.getLogger(__name__)


class DeepPartitionEnsemble(EnsembleClassifier):
    """
    Implementation of Deep Partition Aggregation Defense. Training data is partitioned into
    disjoint buckets based on a hash function and a classifier is trained on each bucket.

    | Paper link: https://arxiv.org/abs/2006.14768
    """

    estimator_params = EnsembleClassifier.estimator_params + [
        "hash_function",
        "ensemble_size",
    ]

    def __init__(
        self,
        classifiers: Union["CLASSIFIER_NEURALNETWORK_TYPE", List["CLASSIFIER_NEURALNETWORK_TYPE"]],
        hash_function: Optional[Callable] = None,
        ensemble_size: int = 50,
        channels_first: bool = False,
        clip_values: Optional["CLIP_VALUES_TYPE"] = None,
        preprocessing_defences: Union["Preprocessor", List["Preprocessor"], None] = None,
        postprocessing_defences: Union["Postprocessor", List["Postprocessor"], None] = None,
        preprocessing: "PREPROCESSING_TYPE" = (0.0, 1.0),
    ) -> None:
        """
        :param classifiers: The base model definition to use for defining the ensemble.
               If a list, the list must be the same size as the ensemble size.
        :param hash_function: The function used to partition the training data. If empty, the hash function
               will use the sum of the input values modulo the ensemble size for partitioning.
        :param ensemble_size: The number of models in the ensemble.
        :param channels_first: Set channels first or last.
        :param clip_values: Tuple of the form `(min, max)` of floats or `np.ndarray` representing the minimum and
               maximum values allowed for features. If floats are provided, these will be used as the range of all
               features. If arrays are provided, each value will be considered the bound for a feature, thus
               the shape of clip values needs to match the total number of features.
        :param preprocessing_defences: Preprocessing defence(s) to be applied by the classifier. Not applicable
               in this classifier.
        :param postprocessing_defences: Postprocessing defence(s) to be applied by the classifier.
        :param preprocessing: Tuple of the form `(subtrahend, divisor)` of floats or `np.ndarray` of values to be
               used for data preprocessing. The first value will be subtracted from the input. The input will then
               be divided by the second one. Not applicable in this classifier.
        """
        self.can_fit = False  # self.fit() cannot be used with models loaded from disk
        if not isinstance(classifiers, list):
            warnings.warn(
                "If a single classifier is passed, it should not have been loaded \
                from disk due to cloning errors with models loaded from disk. If you are \
                using pre-trained model(s), create a list of Estimator objects the same \
                length as the ensemble size"
            )
            self.can_fit = True

            if hasattr(classifiers, "clone_for_refitting"):
                # Initialize the ensemble based on the provided architecture
                # Use ART's cloning if possible
                try:
                    classifiers = [classifiers.clone_for_refitting() for _ in range(ensemble_size)]  # type: ignore
                except ValueError as error:
                    warnings.warn("Switching to deepcopy due to ART Cloning Error: " + str(error))
                    classifiers = [copy.deepcopy(classifiers) for _ in range(ensemble_size)]  # type: ignore
            else:
                classifiers = [copy.deepcopy(classifiers) for _ in range(ensemble_size)]
        elif isinstance(classifiers, list) and len(classifiers) != ensemble_size:
            raise ValueError("The length of the classifier list must be the same as the ensemble size")

        super().__init__(
            classifiers=classifiers,
            clip_values=clip_values,
            channels_first=channels_first,
            preprocessing_defences=preprocessing_defences,
            postprocessing_defences=postprocessing_defences,
            preprocessing=preprocessing,
        )

        if hash_function is None:

            def default_hash(x):
                return int(np.sum(x)) % ensemble_size

            self.hash_function = default_hash
        else:
            self.hash_function = hash_function

        self.ensemble_size = ensemble_size

    def predict(  # pylint: disable=W0221
        self, x: np.ndarray, batch_size: int = 128, raw: bool = False, max_aggregate: bool = True, **kwargs
    ) -> np.ndarray:
        """
        Perform prediction for a batch of inputs. Aggregation will be performed on the prediction from
        each classifier if max_aggregate is True. Otherwise, the probabilities will be summed instead.
        For logits output set max_aggregate=True, as logits are not comparable between models and should
        not be aggregated using a sum.

        :param x: Input samples.
        :param batch_size: Size of batches.
        :param raw: Return the individual classifier raw outputs (not aggregated).
        :param max_aggregate: Aggregate the predicted classes of each classifier if True. If false, aggregation
               is done using a sum. If raw is true, this arg is ignored
        :return: Array of predictions of shape `(nb_inputs, nb_classes)`, or of shape
                 `(nb_classifiers, nb_inputs, nb_classes)` if `raw=True`.
        """

        if raw:
            return super().predict(x, batch_size=batch_size, raw=True, **kwargs)

        # Aggregate based on top-1 prediction from each classifier
        if max_aggregate:
            preds = super().predict(x, batch_size=batch_size, raw=True, **kwargs)
            aggregated_preds = np.zeros_like(preds, shape=preds.shape[1:])  # pylint: disable=E1123
            for i in range(preds.shape[0]):
                aggregated_preds[np.arange(len(aggregated_preds)), np.argmax(preds[i], axis=1)] += 1
            return aggregated_preds

        # Aggregate based on summing predictions from each classifier
        return super().predict(x, batch_size=batch_size, raw=False, **kwargs)

    def fit(  # pylint: disable=W0221
        self,
        x: np.ndarray,
        y: np.ndarray,
        batch_size: int = 128,
        nb_epochs: int = 20,
        train_dict: Dict = None,
        **kwargs
    ) -> None:
        """
        Fit the classifier on the training set `(x, y)`. Each classifier will be trained with the
        same parameters unless train_dict is provided. If train_dict is provided, the model id's
        specified will use the training parameters in train_dict instead.

        :param x: Training data.
        :param y: Target values (class labels) one-hot-encoded of shape (nb_samples, nb_classes).
        :param batch_size: Size of batches.
        :param nb_epochs: Number of epochs to use for training.
        :param train_dict: A dictionary of training args if certain models need specialized arguments.
               The key should be the model's partition id and this will override any default training
               parameters including batch_size and nb_epochs.
        :param kwargs: Dictionary of framework-specific arguments.
        """
        if self.can_fit:
            # First, partition the data using the hash function
            partition_ind = [[] for _ in range(self.ensemble_size)]  # type: List[List[int]]
            for i, p_x in enumerate(x):
                partition_id = int(self.hash_function(p_x))
                partition_ind[partition_id].append(i)

            # Then, train each model on its assigned partition
            for i in range(self.ensemble_size):
                current_x = x[np.array(partition_ind[i])]
                current_y = y[np.array(partition_ind[i])]

                if train_dict is not None and i in train_dict.keys():
                    self.classifiers[i].fit(current_x, current_y, **train_dict[i])
                else:
                    self.classifiers[i].fit(current_x, current_y, batch_size=batch_size, nb_epochs=nb_epochs, **kwargs)
        else:
            warnings.warn("Cannot call fit() for an ensemble of pre-trained classifiers.")



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements Interval bound propagation based layers

| Paper link: https://ieeexplore.ieee.org/document/8418593
"""
from typing import List, Union, Tuple, Optional

import torch
import numpy as np


class PyTorchIntervalDense(torch.nn.Module):
    """
    Class implementing a dense layer for the interval (box) domain.
    """

    def __init__(self, in_features: int, out_features: int):
        super().__init__()

        self.weight = torch.nn.Parameter(
            torch.normal(mean=torch.zeros(out_features, in_features), std=torch.ones(out_features, in_features))
        )
        self.bias = torch.nn.Parameter(torch.normal(mean=torch.zeros(out_features), std=torch.ones(out_features)))

    def __call__(self, x: "torch.Tensor") -> "torch.Tensor":
        return self.forward(x)

    def forward(self, x: "torch.Tensor") -> "torch.Tensor":
        """
        Performs the forward pass of the dense layer in the interval (box) domain.

        :param x: interval representation of the datapoint.
        :return: output of the convolutional layer on x
        """
        center = (x[:, 1] + x[:, 0]) / 2
        radius = (x[:, 1] - x[:, 0]) / 2

        center = torch.matmul(center, torch.transpose(self.weight, 0, 1)) + self.bias
        radius = torch.matmul(radius, torch.abs(torch.transpose(self.weight, 0, 1)))

        center = torch.unsqueeze(center, dim=1)
        radius = torch.unsqueeze(radius, dim=1)

        return torch.cat([center - radius, center + radius], dim=1)

    def concrete_forward(self, x: "torch.Tensor") -> "torch.Tensor":
        """
        Performs the forward pass of the dense layer.

        :param x: concrete input to the convolutional layer.
        :return: output of the convolutional layer on x
        """
        return torch.matmul(x, torch.transpose(self.weight, 0, 1)) + self.bias


class PyTorchIntervalConv2D(torch.nn.Module):
    """
    Class implementing a convolutional layer in the interval/box domain.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple[int, int]],
        input_shape: Tuple[int, ...],
        device: Union[str, "torch.device"],
        stride: Union[int, Tuple[int, int]] = 1,
        padding: Union[int, Tuple[int, int]] = 0,
        dilation: Union[int, Tuple[int, int]] = 0,
        bias: bool = True,
        supplied_input_weights: Union[None, "torch.Tensor"] = None,
        supplied_input_bias: Union[None, "torch.Tensor"] = None,
        to_debug: bool = False,
    ):
        """
        Creates the equivalent dense weights for the specified convolutional layer.

        :param in_channels: Number of input channels
        :param out_channels: Number of output channels
        :param kernel_size: Shape of the convolutional kernel
        :param device: Device to put the weights onto
        :param stride: The convolution's stride
        :param padding: Size of padding to use
        :param dilation: Dilation to apply to the convolution
        :param bias: If to include a bias term
        :param supplied_input_weights: Load in a pre-defined set of convolutional weights with the correct specification
        :param supplied_input_bias: Load in a pre-defined set of convolutional bias with the correct specification
        :param to_debug: Helper parameter to help with debugging.
        """

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.dilation = dilation
        self.stride = stride
        self.device = device
        self.include_bias = bias
        self.cnn: Optional["torch.nn.Conv2d"] = None

        super().__init__()
        self.conv = torch.nn.Conv2d(
            in_channels=1,
            out_channels=out_channels * in_channels,
            kernel_size=kernel_size,
            padding=padding,
            dilation=dilation,
            bias=False,
            stride=stride,
        ).to(device)
        self.bias_to_grad = None

        if bias:
            self.bias_to_grad = torch.nn.Parameter(torch.rand(out_channels).to(device))

        if to_debug:
            self.conv_debug = torch.nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                bias=bias,
                stride=stride,
                padding=padding,
                dilation=dilation,
            ).to(device)

            if isinstance(kernel_size, tuple):
                self.conv.weight = torch.nn.Parameter(
                    torch.reshape(
                        torch.tensor(self.conv_debug.weight.data.cpu().detach().numpy()),
                        (out_channels * in_channels, 1, kernel_size[0], kernel_size[1]),
                    ).to(device)
                )
            else:
                self.conv.weight = torch.nn.Parameter(
                    torch.reshape(
                        torch.tensor(self.conv_debug.weight.data.cpu().detach().numpy()),
                        (out_channels * in_channels, 1, kernel_size, kernel_size),
                    ).to(device)
                )
            if bias and self.conv_debug.bias is not None:
                self.bias_to_grad = torch.nn.Parameter(
                    torch.tensor(self.conv_debug.bias.data.cpu().detach().numpy()).to(device)
                )

        if supplied_input_weights is not None:
            if isinstance(kernel_size, tuple):
                self.conv.weight = torch.nn.Parameter(
                    torch.reshape(
                        supplied_input_weights,
                        (out_channels * in_channels, 1, kernel_size[0], kernel_size[1]),
                    )
                )
            else:
                self.conv.weight = torch.nn.Parameter(
                    torch.reshape(supplied_input_weights, (out_channels * in_channels, 1, kernel_size, kernel_size))
                )

        if supplied_input_bias is not None:
            self.bias_to_grad = torch.nn.Parameter(supplied_input_bias.to(device))

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.input_height = input_shape[2]
        self.input_width = input_shape[3]

        self.output_height: int = 0
        self.output_width: int = 0

        self.dense_weights, self.bias = self.convert_to_dense(device)

        self.dense_weights = self.dense_weights.to(device)
        if self.bias is not None:
            self.bias = self.bias.to(device)

    def re_convert(self, device: Union[str, "torch.device"]) -> None:
        """
        Re converts the weights into a dense equivalent layer.
        Must be called after every backwards if multiple gradients wish to be taken (like for crafting pgd).
        """
        self.dense_weights, self.bias = self.convert_to_dense(device)
        self.dense_weights = self.dense_weights.to(device)
        if self.bias is not None:
            self.bias = self.bias.to(device)

    def convert_to_dense(self, device: Union[str, "torch.device"]) -> Tuple["torch.Tensor", "torch.Tensor"]:
        """
        Converts the initialised convolutional layer into an equivalent dense layer.

        This function was adapted from:
        https://github.com/deepmind/interval-bound-propagation/blob/217a14d12686e08ebb5cfea1f2748cce58a55913/interval_bound_propagation/src/layer_utils.py#L90

        Here, we adapt the tf1 functionality to work with pytorch.

        Original license:

        coding=utf-8

        Copyright 2019 The Interval Bound Propagation Authors.

        Licensed under the Apache License, Version 2.0 (the "License");
        you may not use this file except in compliance with the License.
        You may obtain a copy of the License at

             http://www.apache.org/licenses/LICENSE-2.0

        Unless required by applicable law or agreed to in writing, software
        distributed under the License is distributed on an "AS IS" BASIS,
        WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.

        See the License for the specific language governing permissions and
        limitations under the License.

        :return: The dense weights and bias equivalent to a Conv layer.
        """

        diagonal_input = torch.reshape(
            torch.eye(self.input_height * self.input_width),
            shape=[self.input_height * self.input_width, 1, self.input_height, self.input_width],
        ).to(device)
        conv_output = self.conv(diagonal_input)
        self.output_height = int(conv_output.shape[2])
        self.output_width = int(conv_output.shape[3])

        # conv is of shape (input_height * input_width, out_channels * in_channels, output_height, output_width).
        # Reshape it to (input_height * input_width * output_channels,
        #                output_height * output_width * input_channels).

        weights = torch.reshape(
            conv_output,
            shape=(
                [
                    self.input_height * self.input_width,
                    self.out_channels,
                    self.in_channels,
                    self.output_height,
                    self.output_width,
                ]
            ),
        )
        weights = torch.permute(weights, (2, 0, 1, 3, 4))
        weights = torch.reshape(
            weights,
            shape=(
                [
                    self.input_height * self.input_width * self.in_channels,
                    self.output_height * self.output_width * self.out_channels,
                ]
            ),
        )

        if self.bias_to_grad is not None:
            self.bias = torch.unsqueeze(self.bias_to_grad, dim=-1)
            bias = self.bias.expand(-1, self.output_height * self.output_width)
            bias = bias.flatten()
        else:
            bias = None
        return torch.transpose(weights, 0, 1), bias

    def forward(self, x: "torch.Tensor") -> "torch.Tensor":
        """
        Performs the forward pass of the convolutional layer in the interval (box) domain by using
        the equivalent dense representation.

        :param x: interval representation of the datapoint.
        :return: output of the convolutional layer on x
        """
        x = torch.reshape(x, (x.shape[0], 2, -1))

        center = (x[:, 1] + x[:, 0]) / 2
        radius = (x[:, 1] - x[:, 0]) / 2

        center = torch.matmul(center, torch.transpose(self.dense_weights, 0, 1)) + self.bias
        radius = torch.matmul(radius, torch.abs(torch.transpose(self.dense_weights, 0, 1)))

        center = torch.unsqueeze(center, dim=1)
        radius = torch.unsqueeze(radius, dim=1)

        x = torch.cat([center - radius, center + radius], dim=1)
        return x.reshape((-1, 2, self.out_channels, self.output_height, self.output_width))

    def concrete_forward(self, x: "torch.Tensor") -> "torch.Tensor":
        """
        Performs the forward pass using the equivalent dense representation of the convolutional layer.

        :param x: concrete input to the convolutional layer.
        :return: output of the convolutional layer on x
        """
        x = torch.reshape(x, (x.shape[0], -1))
        if self.bias is None:
            x = torch.matmul(x, torch.transpose(self.dense_weights, 0, 1))
        else:
            x = torch.matmul(x, torch.transpose(self.dense_weights, 0, 1)) + self.bias
        return x.reshape((-1, self.out_channels, self.output_height, self.output_width))

    def conv_forward(self, x: "torch.Tensor") -> "torch.Tensor":
        """
        Method for efficiently interfacing with adversarial attacks.

        Backpropagating through concrete_forward is too slow if adversarial attacks need to be generated on-the fly
        or require a large amount of iterations.

        This method will create a regular conv layer with the right parameters to use.

        :param x: concrete input to the convolutional layer.
        :return: output of the convolutional layer on x
        """
        if self.cnn is None:
            self.cnn = torch.nn.Conv2d(
                in_channels=self.in_channels,
                out_channels=self.out_channels,
                kernel_size=self.kernel_size,
                bias=self.include_bias,
                stride=self.stride,
                padding=self.padding,
                dilation=self.dilation,
            ).to(self.device)
        if isinstance(self.kernel_size, tuple):
            self.cnn.weight.data = torch.reshape(
                torch.tensor(self.conv.weight.data.cpu().detach().numpy()),
                (self.out_channels, self.in_channels, self.kernel_size[0], self.kernel_size[1]),
            ).to(self.device)
        else:
            self.cnn.weight.data = torch.reshape(
                torch.tensor(self.conv.weight.data.cpu().detach().numpy()),
                (self.out_channels, self.in_channels, self.kernel_size, self.kernel_size),
            ).to(self.device)
        if self.cnn.bias is not None and self.bias_to_grad is not None:
            self.cnn.bias.data = torch.tensor(self.bias_to_grad.data.cpu().detach().numpy()).to(self.device)

        if self.cnn is not None:
            return self.cnn(x)
        raise ValueError("The convolutional layer for attack mode was not created properly")


class PyTorchIntervalFlatten(torch.nn.Module):
    """
    Layer to handle flattening on both interval and concrete data
    """

    def __init__(self):
        super().__init__()

    def __call__(self, x: "torch.Tensor") -> "torch.Tensor":
        return self.concrete_forward(x)

    @staticmethod
    def forward(x: "torch.Tensor") -> "torch.Tensor":
        """
        Flattens the provided abstract input

        :param x: datapoint in the interval domain
        :return: Flattened input preserving the batch and bounds dimensions.
        """
        return torch.reshape(x, (x.shape[0], 2, -1))

    @staticmethod
    def concrete_forward(x: "torch.Tensor") -> "torch.Tensor":
        """
        Flattens the provided concrete input

        :param x: datapoint in the concrete domain
        :return: Flattened input preserving the batch dimension.
        """
        return torch.reshape(x, (x.shape[0], -1))


class PyTorchIntervalReLU(torch.nn.Module):
    """
    ReLU activation on both interval and concrete data
    """

    def __init__(self):
        super().__init__()
        self.activation = torch.nn.ReLU()

    def __call__(self, x: "torch.Tensor") -> "torch.Tensor":
        return self.forward(x)

    def forward(self, x: "torch.Tensor") -> "torch.Tensor":
        """
        Abstract pass through the ReLU function

        :param x: abstract input to the activation function.
        :return: abstract outputs from the ReLU.
        """
        return self.activation(x)

    def concrete_forward(self, x: "torch.Tensor") -> "torch.Tensor":
        """
        Concrete pass through the ReLU function

        :param x: concrete input to the activation function.
        :return: concrete outputs from the ReLU.
        """
        return self.activation(x)


class PyTorchIntervalBounds:
    """
    Class providing functionality for computing operations related to interval bounds
    """

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)  # type: ignore

    @staticmethod
    def certify(preds: np.ndarray, labels: np.ndarray) -> np.ndarray:
        """
        Check if the data has been certifiably classified correct.

        :param preds: interval predictions
        :param labels: associated labels (not one-hot encoded).
        :return: array of True or False if predictions are certifiable
        """

        cert_bounds = np.copy(preds[:, 1])  # Take the upper bounds of all the predictions
        for i, label in enumerate(labels):
            cert_bounds[i, label] = preds[i, 0, label]  # Replace the correct prediction with its lower bound
        return np.argmax(cert_bounds, axis=1) == labels

    @staticmethod
    def concrete_to_interval(
        x: np.ndarray,
        bounds: Union[float, List[float], np.ndarray],
        limits: Optional[Union[List[float], np.ndarray]] = None,
    ) -> np.ndarray:
        """
        Helper function converts a datapoint it into its interval representation

        :param x: input datapoint of shape [batch size, feature_1, feature_2, ...]
        :param bounds: Either a scalar to apply to the whole datapoint, or an array of [2, feature_1, feature_2]
                       where bounds[0] are the lower bounds and bounds[1] are the upper bound
        :param limits: if to clip to a range with limits[0] being the lower bounds and limits[1] being upper bounds.
        :return: Data of the form [batch_size, 2, feature_1, feature_2, ...]
                 where [batch_size, 0, x.shape] are the lower bounds and [batch_size, 1, x.shape] are the upper bounds.
        """

        x = np.expand_dims(x, axis=1)

        if isinstance(bounds, float):
            up_x = x + bounds
            lb_x = x - bounds
        elif isinstance(bounds, (list, np.ndarray)):
            up_x = x + bounds[1]
            lb_x = x - bounds[0]
        else:
            raise ValueError("Bounds must be a float, list, or numpy array.")

        final_batched_input = np.concatenate((lb_x, up_x), axis=1)

        if limits is not None:
            return np.clip(final_batched_input, limits[0], limits[1])

        return final_batched_input



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2019
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Implementation of the High-Confidence-Low-Uncertainty (HCLU) adversarial example formulation by Grosse et al. (2018)

| Paper link: https://arxiv.org/abs/1812.02606
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import logging
from typing import Optional

import numpy as np
from scipy.optimize import minimize
from tqdm.auto import trange

from art.attacks.attack import EvasionAttack
from art.estimators.classification.GPy import GPyGaussianProcessClassifier

logger = logging.getLogger(__name__)


class HighConfidenceLowUncertainty(EvasionAttack):
    """
    Implementation of the High-Confidence-Low-Uncertainty (HCLU) adversarial example formulation by Grosse et al. (2018)

    | Paper link: https://arxiv.org/abs/1812.02606
    """

    attack_params = ["conf", "unc_increase", "min_val", "max_val", "verbose"]
    _estimator_requirements = (GPyGaussianProcessClassifier,)

    def __init__(
        self,
        classifier: GPyGaussianProcessClassifier,
        conf: float = 0.95,
        unc_increase: float = 100.0,
        min_val: float = 0.0,
        max_val: float = 1.0,
        verbose: bool = True,
    ) -> None:
        """
        :param classifier: A trained model of type GPYGaussianProcessClassifier.
        :param conf: Confidence that examples should have, if there were to be classified as 1.0 maximally.
        :param unc_increase: Value uncertainty is allowed to deviate, where 1.0 is original value.
        :param min_val: minimal value any feature can take.
        :param max_val: maximal value any feature can take.
        :param verbose: Show progress bars.
        """
        super().__init__(estimator=classifier)
        self.conf = conf
        self.unc_increase = unc_increase
        self.min_val = min_val
        self.max_val = max_val
        self.verbose = verbose
        self._check_params()

    def generate(self, x: np.ndarray, y: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        """
        Generate adversarial examples and return them as an array.

        :param x: An array with the original inputs to be attacked.
        :param y: Target values (class labels) one-hot-encoded of shape (nb_samples, nb_classes) or indices of shape
                  (nb_samples,).
        :return: An array holding the adversarial examples.
        """
        x_adv = copy.copy(x)

        def minfun(x, args):  # minimize L2 norm
            return np.sum(np.sqrt((x - args["orig"]) ** 2))

        def constraint_conf(x, args):  # constraint for confidence
            pred = args["classifier"].predict(x.reshape(1, -1))[0, 0]
            if args["class_zero"]:
                pred = 1.0 - pred
            return (pred - args["conf"]).reshape(-1)

        def constraint_unc(x, args):  # constraint for uncertainty
            cur_unc = (args["classifier"].predict_uncertainty(x.reshape(1, -1))).reshape(-1)
            return (args["max_uncertainty"] - cur_unc)[0]

        bounds = []
        # adding bounds, to not go away from original data
        for i in range(np.shape(x)[1]):
            bounds.append((self.min_val, self.max_val))
        for i in trange(x.shape[0], desc="HCLU", disable=not self.verbose):  # go through data and craft
            # get properties for attack
            max_uncertainty = self.unc_increase * self.estimator.predict_uncertainty(x_adv[i].reshape(1, -1))
            class_zero = not self.estimator.predict(x_adv[i].reshape(1, -1))[0, 0] < 0.5
            init_args = {
                "classifier": self.estimator,
                "class_zero": class_zero,
                "max_uncertainty": max_uncertainty,
                "conf": self.conf,
            }
            constr_conf = {"type": "ineq", "fun": constraint_conf, "args": (init_args,)}
            constr_unc = {"type": "ineq", "fun": constraint_unc, "args": (init_args,)}
            args = {"args": init_args, "orig": x[i].reshape(-1)}
            # finally, run optimization
            x_adv[i] = minimize(
                minfun,
                x_adv[i],
                args=args,
                bounds=bounds,
                constraints=[constr_conf, constr_unc],
            )["x"]

        return x_adv

    def _check_params(self) -> None:

        if self.conf <= 0.5 or self.conf > 1.0:
            raise ValueError("Confidence value has to be a value between 0.5 and 1.0.")

        if self.unc_increase <= 0.0:
            raise ValueError("Value to increase uncertainty must be positive.")

        if self.min_val > self.max_val:
            raise ValueError("Maximum has to be larger than minimum.")

        if not isinstance(self.verbose, bool):
            raise ValueError("The argument `verbose` has to be of type bool.")



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import unittest
import numpy as np

from art.estimators.regression.keras import KerasRegressor

from tests.utils import TestBase, master_seed, get_tabular_regressor_kr

logger = logging.getLogger(__name__)


class TestKerasRegressor(TestBase):
    @classmethod
    def setUpClass(cls):
        master_seed(seed=1234, set_tensorflow=True)
        super().setUpClass()

        import tensorflow as tf

        tf.compat.v1.disable_eager_execution()

        cls.art_model = get_tabular_regressor_kr()

    def test_type(self):
        with self.assertRaises(TypeError):
            KerasRegressor(model="model")

    def test_predict(self):
        y_predicted = self.art_model.predict(self.x_test_diabetes[:4])
        y_expected = np.asarray([[24.9], [52.7], [30.4], [68.1]])
        np.testing.assert_array_almost_equal(y_predicted, y_expected, decimal=1)

    def test_save(self):
        self.art_model.save(filename="test.file", path=None)
        self.art_model.save(filename="test.file", path="./")

    def test_input_shape(self):
        np.testing.assert_equal(self.art_model.input_shape, (10,))

    def test_input_layer(self):
        np.testing.assert_equal(isinstance(self.art_model.input_layer, int), True)

    def test_output_layer(self):
        np.testing.assert_equal(isinstance(self.art_model.output_layer, int), True)

    def test_compute_loss(self):
        test_loss = self.art_model.compute_loss(self.x_test_diabetes[:4], self.y_test_diabetes[:4])
        loss_expected = [6089.8, 2746.3, 5306.8, 1554.9]
        np.testing.assert_array_almost_equal(test_loss, loss_expected, decimal=1)

    def test_loss_gradient(self):
        grad = self.art_model.loss_gradient(self.x_test_diabetes[:4], self.y_test_diabetes[:4])
        grad_expected = [-333.9, 586.4, -1190.9, -123.9, -1206.2, -883.7, 295.9, -830.5, -1333.1, -553.8]
        np.testing.assert_array_almost_equal(grad[0], grad_expected, decimal=1)

    def test_get_activations(self):
        act = self.art_model.get_activations(self.x_test_diabetes[:4], 1)
        act_expected = [0, 0, 0, 7.8, 8.5, 0, 5.6, 0, 6.6, 5.8]
        np.testing.assert_array_almost_equal(act[0], act_expected, decimal=1)


class TestKerasRegressorClass(TestBase):
    @classmethod
    def setUpClass(cls):
        master_seed(seed=1234, set_tensorflow=True)
        super().setUpClass()

        import tensorflow as tf
        import tensorflow.keras as keras

        tf.compat.v1.disable_eager_execution()

        class TestModel(tf.keras.Model):
            def __init__(self):
                super().__init__()
                self.dense1 = keras.layers.Dense(10, activation=tf.nn.relu)
                self.dense2 = keras.layers.Dense(100, activation=tf.nn.relu)
                self.dense3 = keras.layers.Dense(10, activation=tf.nn.relu)
                self.dense4 = keras.layers.Dense(1)

            def call(self, inputs):
                x = self.dense1(inputs)
                return self.dense4(self.dense3(self.dense2(x)))

        cls.keras_model = TestModel()
        cls.keras_model.compile(
            loss=keras.losses.CosineSimilarity(axis=-1, reduction="auto", name="cosine_similarity"),
            optimizer=keras.optimizers.Adam(learning_rate=0.01),
            metrics=["accuracy"],
        )
        cls.keras_model.fit(cls.x_train_diabetes, cls.y_train_diabetes)

        cls.art_model = KerasRegressor(model=cls.keras_model)

    def test_type(self):
        with self.assertRaises(TypeError):
            KerasRegressor(model="model")

    def test_predict(self):
        y_predicted = self.art_model.predict(self.x_test_diabetes[:4])
        np.testing.assert_equal(len(np.unique(y_predicted)), 4)

    def test_save(self):
        self.art_model.save(filename="test.file", path=None)
        self.art_model.save(filename="test.file", path="./")

    def test_input_shape(self):
        np.testing.assert_equal(self.art_model.input_shape, (10,))

    def test_input_layer(self):
        np.testing.assert_equal(isinstance(self.art_model.input_layer, int), True)

    def test_output_layer(self):
        np.testing.assert_equal(isinstance(self.art_model.output_layer, int), True)

    def test_compute_loss(self):
        test_loss = self.art_model.compute_loss(self.x_test_diabetes[:4], self.y_test_diabetes[:4].astype(np.float32))
        # cosine similarity works on vectors, so it returns the same value for each sample
        np.testing.assert_equal(len(np.unique(test_loss)), 1)

    def test_loss_gradient(self):
        grad = self.art_model.loss_gradient(self.x_test_diabetes[:4], self.y_test_diabetes[:4])
        for row in grad:
            if not np.all((row == 0)):
                np.testing.assert_equal(len(np.unique(row)), 10)


class TestKerasRegressorFunctional(TestBase):
    @classmethod
    def setUpClass(cls):
        master_seed(seed=1234, set_tensorflow=True)
        super().setUpClass()

        import tensorflow as tf
        import keras
        from keras.models import Model

        tf.compat.v1.disable_eager_execution()

        def functional():
            in_layer = keras.layers.Input(shape=(10,))
            layer = keras.layers.Dense(100, activation=tf.nn.relu)(in_layer)
            layer = keras.layers.Dense(10, activation=tf.nn.relu)(layer)
            out_layer = keras.layers.Dense(1)(layer)

            model = Model(inputs=[in_layer], outputs=[out_layer])

            model.compile(
                loss=keras.losses.MeanAbsoluteError(),
                optimizer=keras.optimizers.Adam(learning_rate=0.01),
                metrics=["accuracy"],
            )

            return model

        cls.keras_model = functional()
        cls.keras_model.fit(cls.x_train_diabetes, cls.y_train_diabetes)

        cls.art_model = KerasRegressor(model=cls.keras_model)

    def test_type(self):
        with self.assertRaises(TypeError):
            KerasRegressor(model="model")

    def test_predict(self):
        y_predicted = self.art_model.predict(self.x_test_diabetes[:4])
        np.testing.assert_equal(len(np.unique(y_predicted)), 4)

    def test_save(self):
        self.art_model.save(filename="test.file", path=None)
        self.art_model.save(filename="test.file", path="./")

    def test_input_shape(self):
        np.testing.assert_equal(self.art_model.input_shape, (10,))

    def test_input_layer(self):
        np.testing.assert_equal(isinstance(self.art_model.input_layer, int), True)

    def test_output_layer(self):
        np.testing.assert_equal(isinstance(self.art_model.output_layer, int), True)

    def test_compute_loss(self):
        test_loss = self.art_model.compute_loss(self.x_test_diabetes[:4], self.y_test_diabetes[:4].astype(np.float32))
        np.testing.assert_equal(len(np.unique(test_loss)), 4)

    def test_loss_gradient(self):
        grad = self.art_model.loss_gradient(self.x_test_diabetes[:4], self.y_test_diabetes[:4])
        np.testing.assert_equal(len(np.unique(grad[0])), 10)


if __name__ == "__main__":
    unittest.main()



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import pytest

import numpy as np

from art.estimators.classification.blackbox import BlackBoxClassifierNeuralNetwork, BlackBoxClassifier
from tests.utils import ARTTestException


def test_blackbox_existing_predictions(art_warning, get_mnist_dataset):
    try:
        _, (x_test, y_test) = get_mnist_dataset

        limited_x_test = x_test[:500]
        limited_y_test = y_test[:500]

        bb = BlackBoxClassifier((limited_x_test, limited_y_test), (28, 28, 1), 10, clip_values=(0, 255))
        assert np.array_equal(bb.predict(limited_x_test), limited_y_test)

        with pytest.raises(ValueError):
            bb.predict(x_test[:600])

    except ARTTestException as e:
        art_warning(e)


def test_blackbox_existing_predictions_fuzzy(art_warning):
    try:
        x = np.array([0, 3])
        fuzzy_x = np.array([0, 3.00001])
        y = np.array([[1, 0], [0, 1]])
        bb = BlackBoxClassifier((x, y), (1,), 2, fuzzy_float_compare=True)
        assert np.array_equal(bb.predict(fuzzy_x), y)
    except ARTTestException as e:
        art_warning(e)


def test_blackbox_nn_existing_predictions(art_warning, get_mnist_dataset):
    try:
        _, (x_test, y_test) = get_mnist_dataset

        limited_x_test = x_test[:500]
        limited_y_test = y_test[:500]

        bb = BlackBoxClassifierNeuralNetwork((limited_x_test, limited_y_test), (28, 28, 1), 10, clip_values=(0, 255))
        assert np.array_equal(bb.predict(limited_x_test), limited_y_test)

        with pytest.raises(ValueError):
            bb.predict(x_test[:600])

    except ARTTestException as e:
        art_warning(e)


def test_blackbox_nn_existing_predictions_fuzzy(art_warning):
    try:
        x = np.array([0, 3])
        fuzzy_x = np.array([0, 3.00001])
        y = np.array([[1, 0], [0, 1]])
        bb = BlackBoxClassifierNeuralNetwork((x, y), (1,), 2, fuzzy_float_compare=True)
        assert np.array_equal(bb.predict(fuzzy_x), y)
    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
import pytest

import numpy as np

from art.utils import load_dataset
from art.estimators.certification.derandomized_smoothing import (
    PyTorchDeRandomizedSmoothing,
    TensorFlowV2DeRandomizedSmoothing,
)
from tests.utils import ARTTestException


@pytest.fixture()
def fix_get_mnist_data():
    """
    Get the first 128 samples of the mnist test set with channels first format

    :return: First 128 sample/label pairs of the MNIST test dataset.
    """
    nb_test = 128

    (_, _), (x_test, y_test), _, _ = load_dataset("mnist")
    x_test = np.squeeze(x_test).astype(np.float32)
    x_test = np.expand_dims(x_test, axis=1)
    y_test = np.argmax(y_test, axis=1)

    x_test, y_test = x_test[:nb_test], y_test[:nb_test]
    return x_test, y_test


@pytest.fixture()
def fix_get_cifar10_data():
    """
    Get the first 128 samples of the cifar10 test set

    :return: First 128 sample/label pairs of the cifar10 test dataset.
    """
    nb_test = 128

    (_, _), (x_test, y_test), _, _ = load_dataset("cifar10")
    y_test = np.argmax(y_test, axis=1)
    x_test, y_test = x_test[:nb_test], y_test[:nb_test]
    x_test = np.transpose(x_test, (0, 3, 1, 2))  # return in channels first format
    return x_test.astype(np.float32), y_test


@pytest.mark.skip_framework("mxnet", "non_dl_frameworks", "tensorflow1", "keras", "kerastf", "tensorflow2")
def test_pytorch_training(art_warning, fix_get_mnist_data, fix_get_cifar10_data):
    """
    Check that the training loop for pytorch does not result in errors
    """
    import torch
    import torch.optim as optim
    import torch.nn as nn

    device = "cuda" if torch.cuda.is_available() else "cpu"

    class SmallMNISTModel(nn.Module):
        def __init__(self):
            super(SmallMNISTModel, self).__init__()

            self.conv1 = nn.Conv2d(in_channels=2, out_channels=32, kernel_size=(4, 4), dilation=(1, 1), stride=(2, 2))
            self.max_pool = nn.MaxPool2d(2, stride=2)
            self.fc1 = nn.Linear(in_features=1152, out_features=100)

            self.fc2 = nn.Linear(in_features=100, out_features=10)

            self.relu = nn.ReLU()

        def forward(self, x):
            if isinstance(x, np.ndarray):
                x = torch.from_numpy(x).to(device)
            x = self.relu(self.conv1(x))
            x = self.max_pool(x)
            x = torch.flatten(x, 1)
            x = self.relu(self.fc1(x))
            return self.fc2(x)

    class SmallCIFARModel(nn.Module):
        def __init__(self):
            super(SmallCIFARModel, self).__init__()

            self.conv1 = nn.Conv2d(in_channels=6, out_channels=32, kernel_size=(4, 4), dilation=(1, 1), stride=(2, 2))
            self.max_pool = nn.MaxPool2d(2, stride=2)
            self.fc1 = nn.Linear(in_features=1568, out_features=100)

            self.fc2 = nn.Linear(in_features=100, out_features=10)

            self.relu = nn.ReLU()

        def forward(self, x):
            if isinstance(x, np.ndarray):
                x = torch.from_numpy(x).to(device)
            x = self.relu(self.conv1(x))
            x = self.max_pool(x)
            x = torch.flatten(x, 1)
            x = self.relu(self.fc1(x))
            return self.fc2(x)

    for dataset, dataset_name in zip([fix_get_mnist_data, fix_get_cifar10_data], ["mnist", "cifar"]):
        if dataset_name == "mnist":
            ptc = SmallMNISTModel().to(device)
            input_shape = (2, 28, 28)
        else:
            ptc = SmallCIFARModel().to(device)
            input_shape = (6, 32, 32)

        criterion = nn.CrossEntropyLoss()
        optimizer = optim.SGD(ptc.parameters(), lr=0.01, momentum=0.9)
        try:
            for ablation_type in ["column", "row", "block"]:
                classifier = PyTorchDeRandomizedSmoothing(
                    model=ptc,
                    clip_values=(0, 1),
                    loss=criterion,
                    optimizer=optimizer,
                    input_shape=input_shape,
                    nb_classes=10,
                    ablation_type=ablation_type,
                    ablation_size=5,
                    threshold=0.3,
                    logits=True,
                )
                classifier.fit(x=dataset[0], y=dataset[1], nb_epochs=1)
        except ARTTestException as e:
            art_warning(e)


@pytest.mark.skip_framework("mxnet", "non_dl_frameworks", "tensorflow1", "keras", "kerastf", "pytorch")
def test_tf2_training(art_warning, fix_get_mnist_data, fix_get_cifar10_data):
    """
    Check that the training loop for tensorflow2 does not result in errors
    """
    import tensorflow as tf

    def build_model(input_shape):
        img_inputs = tf.keras.Input(shape=input_shape)
        x = tf.keras.layers.Conv2D(filters=32, kernel_size=(4, 4), strides=(2, 2), activation="relu")(img_inputs)
        x = tf.keras.layers.MaxPool2D(pool_size=(2, 2), strides=2)(x)
        # tensorflow uses channels last and we are loading weights from an originally trained pytorch model
        x = tf.transpose(x, (0, 3, 1, 2))
        x = tf.keras.layers.Flatten()(x)
        x = tf.keras.layers.Dense(100, activation="relu")(x)
        x = tf.keras.layers.Dense(10)(x)
        return tf.keras.Model(inputs=img_inputs, outputs=x)

    loss_object = tf.keras.losses.CategoricalCrossentropy(from_logits=True)
    optimizer = tf.keras.optimizers.SGD(learning_rate=0.01)

    for dataset, dataset_name in zip([fix_get_mnist_data, fix_get_cifar10_data], ["mnist", "cifar"]):
        if dataset_name == "mnist":
            input_shape = (28, 28, 2)
        else:
            input_shape = (32, 32, 6)
        net = build_model(input_shape=input_shape)

        try:
            for ablation_type in ["column", "row", "block"]:
                ablation_size = 5
                classifier = TensorFlowV2DeRandomizedSmoothing(
                    model=net,
                    clip_values=(0, 1),
                    loss_object=loss_object,
                    optimizer=optimizer,
                    input_shape=input_shape,
                    nb_classes=10,
                    ablation_type=ablation_type,
                    ablation_size=ablation_size,
                    threshold=0.3,
                    logits=True,
                )
                x = np.transpose(np.copy(dataset[0]), (0, 2, 3, 1))  # put channels last
                classifier.fit(x=x, y=dataset[1], nb_epochs=1)
        except ARTTestException as e:
            art_warning(e)


@pytest.mark.skip_framework("mxnet", "non_dl_frameworks", "tensorflow1", "keras", "kerastf", "tensorflow2")
def test_pytorch_mnist_certification(art_warning, fix_get_mnist_data):
    """
    Assert that the correct number of certifications are given for the MNIST dataset
    """
    import torch
    import torch.optim as optim
    import torch.nn as nn

    device = "cuda" if torch.cuda.is_available() else "cpu"

    class SmallMNISTModel(nn.Module):
        def __init__(self):
            super(SmallMNISTModel, self).__init__()

            self.conv1 = nn.Conv2d(in_channels=2, out_channels=32, kernel_size=(4, 4), dilation=(1, 1), stride=(2, 2))
            self.max_pool = nn.MaxPool2d(2, stride=2)
            self.fc1 = nn.Linear(in_features=1152, out_features=100)

            self.fc2 = nn.Linear(in_features=100, out_features=10)

            self.relu = nn.ReLU()

        def forward(self, x):
            if isinstance(x, np.ndarray):
                x = torch.from_numpy(x).to(device)
            x = self.relu(self.conv1(x))
            x = self.max_pool(x)
            x = torch.flatten(x, 1)
            x = self.relu(self.fc1(x))
            return self.fc2(x)

        def load_weights(self):

            fpath = os.path.join(
                os.path.dirname(os.path.dirname(__file__)), "../../utils/resources/models/certification/derandomized/"
            )
            self.conv1.weight = nn.Parameter(torch.from_numpy(np.load(fpath + "W_CONV2D1_MNIST.npy")).float())
            self.conv1.bias = nn.Parameter(torch.from_numpy(np.load(fpath + "B_CONV2D1_MNIST.npy")).float())

            self.fc1.weight = nn.Parameter(torch.from_numpy(np.load(fpath + "W_DENSE1_MNIST.npy")).float())
            self.fc1.bias = nn.Parameter(torch.from_numpy(np.load(fpath + "B_DENSE1_MNIST.npy")).float())

            self.fc2.weight = nn.Parameter(torch.from_numpy(np.load(fpath + "W_DENSE2_MNIST.npy")).float())
            self.fc2.bias = nn.Parameter(torch.from_numpy(np.load(fpath + "B_DENSE2_MNIST.npy")).float())

    ptc = SmallMNISTModel()
    ptc.load_weights()
    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(ptc.parameters(), lr=0.01)

    try:
        for ablation_type in ["column", "block"]:
            if ablation_type == "column":
                size_to_certify = 5
                ablation_size = 2
            else:
                """
                the model was trained on column ablations, so make the block task simpler so that a
                degree of certification is obtained.
                """
                size_to_certify = 1
                ablation_size = 5

            classifier = PyTorchDeRandomizedSmoothing(
                model=ptc,
                clip_values=(0, 1),
                loss=criterion,
                optimizer=optimizer,
                input_shape=(2, 28, 28),
                nb_classes=10,
                ablation_type=ablation_type,
                ablation_size=ablation_size,
                threshold=0.3,
                logits=True,
            )

            preds = classifier.predict(np.copy(fix_get_mnist_data[0]))
            num_certified = classifier.ablator.certify(preds, size_to_certify=size_to_certify)

            if ablation_type == "column":
                assert np.sum(num_certified) == 52
            else:
                assert np.sum(num_certified) == 22
    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("mxnet", "non_dl_frameworks", "tensorflow1", "keras", "kerastf", "pytorch")
def test_tf2_mnist_certification(art_warning, fix_get_mnist_data):
    """
    Assert that the correct number of certifications are given for the MNIST dataset
    """

    import tensorflow as tf

    def build_model(input_shape):
        img_inputs = tf.keras.Input(shape=input_shape)
        x = tf.keras.layers.Conv2D(filters=32, kernel_size=(4, 4), strides=(2, 2), activation="relu")(img_inputs)
        x = tf.keras.layers.MaxPool2D(pool_size=(2, 2), strides=2)(x)
        # tensorflow uses channels last and we are loading weights from an originally trained pytorch model
        x = tf.transpose(x, (0, 3, 1, 2))
        x = tf.keras.layers.Flatten()(x)
        x = tf.keras.layers.Dense(100, activation="relu")(x)
        x = tf.keras.layers.Dense(10)(x)
        return tf.keras.Model(inputs=img_inputs, outputs=x)

    def get_weights():
        fpath = os.path.join(
            os.path.dirname(os.path.dirname(__file__)), "../../utils/resources/models/certification/derandomized/"
        )
        weight_names = [
            "W_CONV2D1_MNIST.npy",
            "B_CONV2D1_MNIST.npy",
            "W_DENSE1_MNIST.npy",
            "B_DENSE1_MNIST.npy",
            "W_DENSE2_MNIST.npy",
            "B_DENSE2_MNIST.npy",
        ]
        weight_list = []
        for name in weight_names:
            w = np.load(fpath + name)
            if "W_CONV" in name:
                w = np.transpose(w, (2, 3, 1, 0))
            if "W_DENSE" in name:
                w = np.transpose(w)
            weight_list.append(w)
        return weight_list

    net = build_model(input_shape=(28, 28, 2))
    net.set_weights(get_weights())

    loss_object = tf.keras.losses.CategoricalCrossentropy(from_logits=True)
    optimizer = tf.keras.optimizers.SGD(learning_rate=0.01)

    try:
        for ablation_type in ["column", "block"]:
            if ablation_type == "column":
                size_to_certify = 5
                ablation_size = 2
            else:
                """
                the model was trained on column ablations, so make the block task simpler so that a
                degree of certification is obtained.
                """
                size_to_certify = 1
                ablation_size = 5

            classifier = TensorFlowV2DeRandomizedSmoothing(
                model=net,
                clip_values=(0, 1),
                loss_object=loss_object,
                optimizer=optimizer,
                input_shape=(28, 28, 2),
                nb_classes=10,
                ablation_type=ablation_type,
                ablation_size=ablation_size,
                threshold=0.3,
                logits=True,
            )

            x = np.copy(fix_get_mnist_data[0])
            x = np.squeeze(x)
            x = np.expand_dims(x, axis=-1)
            preds = classifier.predict(x)
            num_certified = classifier.ablator.certify(preds, size_to_certify=size_to_certify)

            if ablation_type == "column":
                assert np.sum(num_certified) == 52
            else:
                assert np.sum(num_certified) == 22

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import numpy as np
import pytest

from art.attacks.poisoning import HiddenTriggerBackdoor
from art.attacks.poisoning import PoisoningAttackBackdoor
from art.attacks.poisoning.perturbations import add_pattern_bd
from art.estimators.classification.pytorch import PyTorchClassifier

from tests.utils import ARTTestException

logger = logging.getLogger(__name__)


@pytest.mark.skip_framework("non_dl_frameworks", "tensorflow1", "tensorflow2v1", "mxnet")
def test_poison(art_warning, get_default_mnist_subset, image_dl_estimator):
    try:
        (x_train, y_train), (_, _) = get_default_mnist_subset
        classifier, _ = image_dl_estimator(functional=True)

        if isinstance(classifier, PyTorchClassifier):

            def mod(x):
                original_dtype = x.dtype
                x = np.transpose(x, (0, 2, 3, 1)).astype(np.float32)
                x = add_pattern_bd(x)
                x = np.transpose(x, (0, 3, 1, 2)).astype(np.float32)
                return x.astype(original_dtype)

        else:

            def mod(x):
                original_dtype = x.dtype
                x = add_pattern_bd(x)
                return x.astype(original_dtype)

        backdoor = PoisoningAttackBackdoor(mod)
        target = y_train[0]
        diff_index = list(set(np.arange(len(y_train))) - set(np.where(np.all(y_train == target, axis=1))[0]))[0]
        source = y_train[diff_index]
        attack = HiddenTriggerBackdoor(
            classifier,
            eps=0.3,
            target=target,
            source=source,
            feature_layer=len(classifier.layer_names) - 2,
            backdoor=backdoor,
            decay_coeff=0.95,
            decay_iter=1,
            max_iter=2,
            batch_size=1,
        )
        poison_data, poison_inds = attack.poison(x_train, y_train)

        with pytest.raises(AssertionError):
            np.testing.assert_equal(poison_data, x_train[poison_inds])

    except ARTTestException as e:
        art_warning(e)


@pytest.mark.skip_framework("non_dl_frameworks", "tensorflow1", "tensorflow2v1", "mxnet")
def test_check_params(art_warning, get_default_mnist_subset, image_dl_estimator):
    try:
        (x_train, y_train), (_, _) = get_default_mnist_subset
        classifier, _ = image_dl_estimator(functional=True)

        if isinstance(classifier, PyTorchClassifier):

            def mod(x):
                original_dtype = x.dtype
                x = np.transpose(x, (0, 2, 3, 1)).astype(np.float32)
                x = add_pattern_bd(x)
                x = np.transpose(x, (0, 3, 1, 2)).astype(np.float32)
                return x.astype(original_dtype)

        else:

            def mod(x):
                original_dtype = x.dtype
                x = add_pattern_bd(x)
                return x.astype(original_dtype)

        backdoor = PoisoningAttackBackdoor(mod)
        target = y_train[0]
        diff_index = list(set(np.arange(len(y_train))) - set(np.where(np.all(y_train == target, axis=1))[0]))[0]
        source = y_train[diff_index]

        # Test non-array target
        with pytest.raises(ValueError):
            _ = HiddenTriggerBackdoor(
                classifier,
                eps=0.3,
                target=0,
                source=source,
                feature_layer=len(classifier.layer_names) - 2,
                backdoor=backdoor,
                decay_coeff=0.95,
                decay_iter=1,
                max_iter=2,
                batch_size=1,
                poison_percent=0.1,
                learning_rate=-1,
            )
        # Test negative LR
        with pytest.raises(ValueError):
            _ = HiddenTriggerBackdoor(
                classifier,
                eps=0.3,
                target=target,
                source=source,
                feature_layer=len(classifier.layer_names) - 2,
                backdoor=backdoor,
                decay_coeff=0.95,
                decay_iter=1,
                max_iter=2,
                batch_size=1,
                poison_percent=0.1,
                learning_rate=-1,
            )
        # Test same target/source
        with pytest.raises(ValueError):
            _ = HiddenTriggerBackdoor(
                classifier,
                eps=0.3,
                target=source,
                source=source,
                feature_layer=len(classifier.layer_names) - 2,
                backdoor=backdoor,
                decay_coeff=0.95,
                decay_iter=1,
                max_iter=2,
                batch_size=1,
                poison_percent=0.1,
            )
        # Test Bad Backdoor type
        with pytest.raises(TypeError):
            _ = HiddenTriggerBackdoor(
                classifier,
                eps=0.3,
                target=target,
                source=source,
                feature_layer=len(classifier.layer_names) - 2,
                backdoor=source,
                decay_coeff=0.95,
                decay_iter=1,
                max_iter=2,
                batch_size=1,
                poison_percent=0.1,
            )
        # Test eps
        with pytest.raises(ValueError):
            _ = HiddenTriggerBackdoor(
                classifier,
                eps=-1,
                target=target,
                source=source,
                feature_layer=len(classifier.layer_names) - 2,
                backdoor=backdoor,
                decay_coeff=0.95,
                decay_iter=1,
                max_iter=2,
                batch_size=1,
                poison_percent=0.1,
            )
        # Test bad feature layer
        with pytest.raises(TypeError):
            _ = HiddenTriggerBackdoor(
                classifier,
                eps=0.3,
                target=target,
                source=source,
                feature_layer=2.5,
                backdoor=backdoor,
                decay_coeff=0.95,
                decay_iter=1,
                max_iter=2,
                batch_size=1,
                poison_percent=0.1,
            )
        # Test negative feature layer
        with pytest.raises(ValueError):
            _ = HiddenTriggerBackdoor(
                classifier,
                eps=0.3,
                target=target,
                source=source,
                feature_layer=-1,
                backdoor=backdoor,
                decay_coeff=0.95,
                decay_iter=1,
                max_iter=2,
                batch_size=1,
                poison_percent=0.1,
            )
        # Test negative decay
        with pytest.raises(ValueError):
            _ = HiddenTriggerBackdoor(
                classifier,
                eps=0.3,
                target=target,
                source=source,
                feature_layer=len(classifier.layer_names) - 2,
                backdoor=backdoor,
                decay_coeff=-1,
                decay_iter=1,
                max_iter=2,
                batch_size=1,
                poison_percent=0.1,
            )
        # Test invalid poison_percent
        with pytest.raises(ValueError):
            _ = HiddenTriggerBackdoor(
                classifier,
                eps=0.3,
                target=target,
                source=source,
                feature_layer=len(classifier.layer_names) - 2,
                backdoor=backdoor,
                decay_coeff=0.95,
                decay_iter=1,
                max_iter=2,
                batch_size=1,
                poison_percent=1.1,
            )

    except ARTTestException as e:
        art_warning(e)



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the abstract estimator for scikit-learn models.
"""
import logging
from typing import Optional, Tuple

from art.estimators.estimator import BaseEstimator

logger = logging.getLogger(__name__)


class ScikitlearnEstimator(BaseEstimator):
    """
    Estimator class for scikit-learn models.
    """

    def _get_input_shape(self, model) -> Optional[Tuple[int, ...]]:
        _input_shape: Optional[Tuple[int, ...]]
        if hasattr(model, "n_features_"):
            _input_shape = (model.n_features_,)
        elif hasattr(model, "n_features_in_"):
            _input_shape = (model.n_features_in_,)
        elif hasattr(model, "feature_importances_"):
            _input_shape = (len(model.feature_importances_),)
        elif hasattr(model, "coef_"):
            if len(model.coef_.shape) == 1:
                _input_shape = (model.coef_.shape[0],)
            else:
                _input_shape = (model.coef_.shape[1],)
        elif hasattr(model, "support_vectors_"):
            _input_shape = (model.support_vectors_.shape[1],)
        elif hasattr(model, "steps"):
            _input_shape = self._get_input_shape(model.steps[0][1])
        else:
            logger.warning("Input shape not recognised. The model might not have been fitted.")
            _input_shape = None
        return _input_shape



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the Mixup data augmentation defence in PyTorch.

| Paper link: https://arxiv.org/abs/1710.09412

| Please keep in mind the limitations of defences. For more information on the limitations of this defence,
    see https://arxiv.org/abs/1803.09868 . For details on how to evaluate classifier security in general, see
    https://arxiv.org/abs/1902.06705
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import Optional, Tuple, TYPE_CHECKING

import numpy as np

from art.defences.preprocessor.preprocessor import PreprocessorPyTorch

if TYPE_CHECKING:
    # pylint: disable=C0412
    import torch

logger = logging.getLogger(__name__)


class MixupPyTorch(PreprocessorPyTorch):
    """
    Implement the Mixup data augmentation defence approach in PyTorch.

    | Paper link: https://arxiv.org/abs/1710.09412

    | Please keep in mind the limitations of defences. For more information on the limitations of this defence,
        see https://arxiv.org/abs/1803.09868 . For details on how to evaluate classifier security in general, see
        https://arxiv.org/abs/1902.06705
    """

    params = ["num_classes", "alpha", "num_mix"]

    def __init__(
        self,
        num_classes: int,
        alpha: float = 1.0,
        num_mix: int = 2,
        apply_fit: bool = True,
        apply_predict: bool = False,
        device_type: str = "gpu",
    ) -> None:
        """
        Create an instance of a Mixup data augmentation object.

        :param num_classes: The number of classes used for one-hot encoding.
        :param alpha: The hyperparameter for the mixing interpolation strength.
        :param num_mix: The number of samples to mix for k-way Mixup.
        :param apply_fit: True if applied during fitting/training.
        :param apply_predict: True if applied during predicting.
        :param device_type: Type of device on which the classifier is run, either `gpu` or `cpu`.
        """
        super().__init__(
            device_type=device_type,
            is_fitted=True,
            apply_fit=apply_fit,
            apply_predict=apply_predict,
        )
        self.num_classes = num_classes
        self.alpha = alpha
        self.num_mix = num_mix
        self._check_params()

    def forward(
        self, x: "torch.Tensor", y: Optional["torch.Tensor"] = None
    ) -> Tuple["torch.Tensor", Optional["torch.Tensor"]]:
        """
        Apply Mixup data augmentation to feature data `x` and labels `y`.

        :param x: Feature data to augment with shape `(batch_size, ...)`.
        :param y: Labels of `x` either one-hot or multi-hot encoded of shape `(nb_samples, nb_classes)`
                  or class indices of shape `(nb_samples,)`.
        :return: Data augmented sample. The returned labels will be probability vectors of shape
                 `(nb_samples, nb_classes)`.
        :raises `ValueError`: If no labels are provided.
        """
        import torch  # lgtm [py/repeated-import]

        if y is None:
            raise ValueError("Labels `y` cannot be None.")

        # convert labels to one-hot encoding
        if len(y.shape) == 2:
            y_one_hot = y
        elif len(y.shape) == 1:
            y_one_hot = torch.nn.functional.one_hot(y, self.num_classes)
        else:
            raise ValueError(
                "Shape of labels not recognised. "
                "Please provide labels in shape (nb_samples,) or (nb_samples, nb_classes)"
            )

        n = x.shape[0]

        # sample the mixing factor from the Dirichlet distribution
        lmbs = np.random.dirichlet([self.alpha] * self.num_mix)

        x_aug = lmbs[0] * x
        y_aug = lmbs[0] * y_one_hot
        for lmb in lmbs[1:]:
            # randomly choose indices for samples to mix
            indices = torch.randperm(n, device=x.device)
            x_aug = x_aug + lmb * x[indices]
            y_aug = y_aug + lmb * y_one_hot[indices]

        return x_aug, y_aug

    def _check_params(self) -> None:
        if self.num_classes <= 0:
            raise ValueError("The number of classes must be positive")

        if self.alpha <= 0:
            raise ValueError("The mixing interpolation strength must be positive.")

        if self.num_mix < 2:
            raise ValueError("The number of samples to mix must be at least 2.")



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2022
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This module implements the Cutout data augmentation defence in TensorFlow.

| Paper link: https://arxiv.org/abs/1708.04552

| Please keep in mind the limitations of defences. For more information on the limitations of this defence,
    see https://arxiv.org/abs/1803.09868 . For details on how to evaluate classifier security in general, see
    https://arxiv.org/abs/1902.06705
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
from typing import Optional, Tuple, TYPE_CHECKING

from art.defences.preprocessor.preprocessor import PreprocessorTensorFlowV2

if TYPE_CHECKING:
    # pylint: disable=C0412
    import tensorflow as tf

logger = logging.getLogger(__name__)


class CutoutTensorFlowV2(PreprocessorTensorFlowV2):
    """
    Implement the Cutout data augmentation defence approach in TensorFlow v2.

    | Paper link: https://arxiv.org/abs/1708.04552

    | Please keep in mind the limitations of defences. For more information on the limitations of this defence,
        see https://arxiv.org/abs/1803.09868 . For details on how to evaluate classifier security in general, see
        https://arxiv.org/abs/1902.06705
    """

    params = ["length", "channels_first", "verbose"]

    def __init__(
        self,
        length: int,
        channels_first: bool = False,
        apply_fit: bool = True,
        apply_predict: bool = False,
        verbose: bool = False,
    ):
        """
        Create an instance of a Cutout data augmentation object.

        :param length: Maximum length of the bounding box.
        :param channels_first: Set channels first or last.
        :param apply_fit: True if applied during fitting/training.
        :param apply_predict: True if applied during predicting.
        :param verbose: Show progress bars.
        """
        super().__init__(is_fitted=True, apply_fit=apply_fit, apply_predict=apply_predict)
        self.length = length
        self.channels_first = channels_first
        self.verbose = verbose
        self._check_params()

    def forward(self, x: "tf.Tensor", y: Optional["tf.Tensor"] = None) -> Tuple["tf.Tensor", Optional["tf.Tensor"]]:
        """
        Apply Cutout data augmentation to sample `x`.

        :param x: Sample to cut out with shape of `NCHW`, `NHWC`, `NCFHW` or `NFHWC`.
                  `x` values are expected to be in the data range [0, 1] or [0, 255].
        :param y: Labels of the sample `x`. This function does not affect them in any way.
        :return: Data augmented sample.
        """
        import tensorflow as tf  # lgtm [py/repeated-import]
        import tensorflow_addons as tfa

        x_ndim = len(x.shape)

        # NCHW/NCFHW/NFHWC --> NHWC
        if x_ndim == 4:
            if self.channels_first:
                # NCHW --> NHWC
                x_nhwc = tf.transpose(x, (0, 2, 3, 1))
            else:
                # NHWC
                x_nhwc = x
        elif x_ndim == 5:
            if self.channels_first:
                # NCFHW --> NFHWC --> NHWC
                nb_clips, channels, clip_size, height, width = x.shape
                x_nfhwc = tf.transpose(x, (0, 2, 3, 4, 1))
                x_nhwc = tf.reshape(x_nfhwc, (nb_clips * clip_size, height, width, channels))
            else:
                # NFHWC --> NHWC
                nb_clips, clip_size, height, width, channels = x.shape
                x_nhwc = tf.reshape(x, (nb_clips * clip_size, height, width, channels))
        else:
            raise ValueError("Unrecognized input dimension. Cutout can only be applied to image and video data.")

        # round down length to be divisible by 2
        length = self.length if self.length % 2 == 0 else max(self.length - 1, 2)

        # apply random cutout
        x_nhwc = tfa.image.random_cutout(x_nhwc, (length, length))

        # NCHW/NCFHW/NFHWC <-- NHWC
        if x_ndim == 4:
            if self.channels_first:
                # NHWC <-- NCHW
                x_aug = tf.transpose(x_nhwc, (0, 3, 1, 2))
            else:
                # NHWC
                x_aug = x_nhwc
        elif x_ndim == 5:  # lgtm [py/redundant-comparison]
            if self.channels_first:
                # NCFHW <-- NFHWC <-- NHWC
                x_nfhwc = tf.reshape(x_nhwc, (nb_clips, clip_size, height, width, channels))
                x_aug = tf.transpose(x_nfhwc, (0, 4, 1, 2, 3))
            else:
                # NFHWC <-- NHWC
                x_aug = tf.reshape(x_nhwc, (nb_clips, clip_size, height, width, channels))

        return x_aug, y

    def _check_params(self) -> None:
        if self.length <= 0:
            raise ValueError("Bounding box length must be positive.")



# MIT License
#
# Copyright (C) The Adversarial Robustness Toolbox (ART) Authors 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import numpy as np
import pytest
import tensorflow as tf

from tests.utils import ARTTestException, master_seed

logger = logging.getLogger(__name__)


@pytest.mark.skip_module("object_detection")
@pytest.mark.skip_framework("pytorch", "tensorflow2v1", "keras", "kerastf", "mxnet", "non_dl_frameworks")
def test_tf_faster_rcnn(art_warning, get_mnist_dataset):
    try:
        master_seed(seed=1234, set_tensorflow=True)

        # Only import if object detection module is available
        from art.estimators.object_detection.tensorflow_faster_rcnn import TensorFlowFasterRCNN

        # Define object detector
        images = tf.placeholder(tf.float32, shape=[1, 28, 28, 1])
        obj_dec = TensorFlowFasterRCNN(images=images)

        # Get test data
        (_, _), (x_test_mnist, y_test_mnist) = get_mnist_dataset
        x_test_mnist = x_test_mnist[:1]

        # First test predict
        result = obj_dec.predict(x_test_mnist)

        assert list(result[0].keys()) == ["boxes", "labels", "scores"]

        assert result[0]["boxes"].shape == (300, 4)
        expected_detection_boxes = np.asarray([0.008862, 0.003788, 0.070454, 0.175931])
        np.testing.assert_array_almost_equal(result[0]["boxes"][2, :], expected_detection_boxes, decimal=3)

        assert result[0]["scores"].shape == (300,)
        expected_detection_scores = np.asarray(
            [
                2.196349e-04,
                7.968055e-05,
                7.811916e-05,
                7.334248e-05,
                6.868376e-05,
                6.861838e-05,
                6.756858e-05,
                6.331169e-05,
                6.313509e-05,
                6.222352e-05,
            ]
        )
        np.testing.assert_array_almost_equal(result[0]["scores"][:10], expected_detection_scores, decimal=3)

        assert result[0]["labels"].shape == (300,)
        # expected_detection_classes = np.asarray([37, 15, 15, 66, 15, 15, 15, 63, 2, 66])
        # np.testing.assert_array_almost_equal(result[0]["labels"][:10], expected_detection_classes)

        # Then test loss gradient
        # Create labels
        y = [{"boxes": result[0]["boxes"], "labels": result[0]["labels"], "scores": np.ones_like(result[0]["labels"])}]

        # Compute gradients
        grads = obj_dec.loss_gradient(x_test_mnist[:1], y)

        assert grads.shape == (1, 28, 28, 1)

        expected_gradients = np.asarray(
            [
                [-0.00298723],
                [-0.0039893],
                [-0.00036253],
                [0.01038542],
                [0.01455704],
                [0.00995643],
                [0.00424966],
                [0.00470569],
                [0.00666382],
                [0.0028694],
                [0.00525351],
                [0.00889174],
                [0.0071413],
                [0.00618231],
                [0.00598106],
                [0.0072665],
                [0.00708815],
                [0.00286943],
                [0.00411595],
                [0.00788978],
                [0.00587319],
                [0.00808631],
                [0.01018151],
                [0.00867905],
                [0.00820272],
                [0.00124911],
                [-0.0042593],
                [0.02380728],
            ]
        )
        np.testing.assert_array_almost_equal(grads[0, 0, :, :], expected_gradients, decimal=2)

        # Then test loss gradient with standard format
        # Create labels
        result_tf = obj_dec.predict(x_test_mnist, standardise_output=False)
        result = obj_dec.predict(x_test_mnist, standardise_output=True)

        from art.estimators.object_detection.utils import convert_tf_to_pt

        result_pt = convert_tf_to_pt(y=result_tf, height=x_test_mnist.shape[1], width=x_test_mnist.shape[2])

        np.testing.assert_array_equal(result[0]["boxes"], result_pt[0]["boxes"])
        np.testing.assert_array_equal(result[0]["labels"], result_pt[0]["labels"])
        np.testing.assert_array_equal(result[0]["scores"], result_pt[0]["scores"])

        y = [{"boxes": result[0]["boxes"], "labels": result[0]["labels"], "scores": np.ones_like(result[0]["labels"])}]

        # Compute gradients
        grads = obj_dec.loss_gradient(x_test_mnist[:1], y, standardise_output=True)

        assert grads.shape == (1, 28, 28, 1)

        expected_gradients = np.asarray(
            [
                [-0.00095965],
                [-0.00265362],
                [-0.00031886],
                [0.01132964],
                [0.01674244],
                [0.01262039],
                [0.0063345],
                [0.00673249],
                [0.00618648],
                [0.00422678],
                [0.00542425],
                [0.00814896],
                [0.00919153],
                [0.01068758],
                [0.00929435],
                [0.00877143],
                [0.00747379],
                [0.0050377],
                [0.00656254],
                [0.00799547],
                [0.0051057],
                [0.00714598],
                [0.01090685],
                [0.00787637],
                [0.00709959],
                [0.00047201],
                [-0.00460457],
                [0.02629307],
            ]
        )
        np.testing.assert_array_almost_equal(grads[0, 0, :, :], expected_gradients, decimal=2)

        obj_dec._sess.close()
        tf.reset_default_graph()

    except ARTTestException as e:
        art_warning(e)


def test_errors(art_warning, get_mnist_dataset):
    try:
        from art.estimators.object_detection.tensorflow_faster_rcnn import TensorFlowFasterRCNN

        images = tf.placeholder(tf.float32, shape=[1, 28, 28, 1])

        with pytest.raises(ValueError):
            TensorFlowFasterRCNN(
                images=images,
                clip_values=(1, 2),
                attack_losses=("loss_classifier", "loss_box_reg", "loss_objectness", "loss_rpn_box_reg"),
            )

        with pytest.raises(ValueError):
            TensorFlowFasterRCNN(
                images=images,
                clip_values=(-1, 1),
                attack_losses=("loss_classifier", "loss_box_reg", "loss_objectness", "loss_rpn_box_reg"),
            )

        from art.defences.postprocessor.rounded import Rounded

        post_def = Rounded()
        with pytest.raises(ValueError):
            TensorFlowFasterRCNN(
                images=images,
                clip_values=(0, 1),
                attack_losses=("loss_classifier", "loss_box_reg", "loss_objectness", "loss_rpn_box_reg"),
                postprocessing_defences=post_def,
            )

    except ARTTestException as e:
        art_warning(e)


def test_preprocessing_defences(art_warning, get_mnist_dataset):
    try:
        from art.estimators.object_detection.tensorflow_faster_rcnn import TensorFlowFasterRCNN
        from art.defences.preprocessor.spatial_smoothing import SpatialSmoothing

        images = tf.placeholder(tf.float32, shape=[1, 28, 28, 3])

        pre_def = SpatialSmoothing()

        with pytest.raises(ValueError):
            _ = TensorFlowFasterRCNN(
                images=images,
                clip_values=(0, 1),
                attack_losses=("loss_classifier", "loss_box_reg", "loss_objectness", "loss_rpn_box_reg"),
                preprocessing_defences=pre_def,
            )

    except ARTTestException as e:
        art_warning(e)


def test_compute_losses(art_warning, get_mnist_dataset):
    try:
        from art.estimators.object_detection.tensorflow_faster_rcnn import TensorFlowFasterRCNN

        images = tf.placeholder(tf.float32, shape=[2, 28, 28, 3])

        # Get test data
        (_, _), (x_test_mnist, y_test_mnist) = get_mnist_dataset

        frcnn = TensorFlowFasterRCNN(
            images=images,
            clip_values=(0, 1),
            attack_losses=(
                "Loss/RPNLoss/localization_loss",
                "Loss/RPNLoss/objectness_loss",
                "Loss/BoxClassifierLoss/localization_loss",
                "Loss/BoxClassifierLoss/classification_loss",
            ),
        )

        # Create labels
        result = frcnn.predict(np.repeat(x_test_mnist[:2].astype(np.float32), repeats=3, axis=3))

        y = [
            {
                "boxes": result[0]["boxes"],
                "labels": result[0]["labels"],
                "scores": np.ones_like(result[0]["labels"]),
            },
            {
                "boxes": result[1]["boxes"],
                "labels": result[1]["labels"],
                "scores": np.ones_like(result[1]["labels"]),
            },
        ]

        # Compute losses
        losses = frcnn.compute_losses(np.repeat(x_test_mnist[:2].astype(np.float32), repeats=3, axis=3), y)

        assert len(losses) == 4

    except ARTTestException as e:
        art_warning(e)


def test_compute_loss(art_warning, get_mnist_dataset):
    try:
        from art.estimators.object_detection.tensorflow_faster_rcnn import TensorFlowFasterRCNN

        images = tf.placeholder(tf.float32, shape=[2, 28, 28, 3])

        # Get test data
        (_, _), (x_test_mnist, y_test_mnist) = get_mnist_dataset

        frcnn = TensorFlowFasterRCNN(
            images=images,
            clip_values=(0, 1),
            attack_losses=(
                "Loss/RPNLoss/localization_loss",
                "Loss/RPNLoss/objectness_loss",
                "Loss/BoxClassifierLoss/localization_loss",
                "Loss/BoxClassifierLoss/classification_loss",
            ),
        )

        # Create labels
        result = frcnn.predict(np.repeat(x_test_mnist[:2].astype(np.float32), repeats=3, axis=3))

        y = [
            {
                "boxes": result[0]["boxes"],
                "labels": result[0]["labels"],
                "scores": np.ones_like(result[0]["labels"]),
            },
            {
                "boxes": result[1]["boxes"],
                "labels": result[1]["labels"],
                "scores": np.ones_like(result[1]["labels"]),
            },
        ]

        # Compute loss
        loss = frcnn.compute_loss(np.repeat(x_test_mnist[:2].astype(np.float32), repeats=3, axis=3), y)

        assert float(loss) == pytest.approx(6.592308044433594, abs=0.5)

    except ARTTestException as e:
        art_warning(e)
